# Allows nested asyncio event loop.
# This library depends on completing tasks on asyncio event loop, which
# will create a problem if the environment, such as web servers,
# GUI applications, Jupyter notebooks, etc, is already running an event loop.
# Using nest_asyncio below fixes this problem.
import nest_asyncio
nest_asyncio.apply()

from .Broker import Broker
from .Strategy import Strategy
from .Bot import Bot
from .Notifier import Notifier, addMessage
from .Data import Data
from .KrakenData import KrakenData
from .CSVData import CSVData
from .BacktestBroker import BacktestBroker
from .LNMBroker import LNMBroker



import time
import pandas as pd
from abc import ABCMeta, abstractmethod


class Strategy(metaclass=ABCMeta):

    def __init__(self, broker=None, **params):
        self.params = params
        self.broker = broker
        self.datas = []
        self.dataIndex = []

        self.highest = 0.0
        self.drawdown = 0.0
        self.maxdrawdown = 0.0
        if broker is not None:
            self.initialCapital = broker.balance
        self.startTime = time.perf_counter()
        self.position = 0
        self.orderBook = pd.DataFrame(columns=[
            'Type',
            'Limit',
            'Quantity',
            'Leverage',
            'Stoploss',
            'Takeprofit',
            'Parent',
            'Strategy',
            'StopExecuted',
            'ProfitExecuted',
        ], index=pd.Series([]))

        self.init()

    @abstractmethod
    def init(self):
        """
        This method is called only once, before any price bar is recieved.
        """
        pass

    @abstractmethod
    def execute(self, datas):
        """
        This method is called for every new bar recieved.
        """
        pass

    def processData(self, datas):
        self.computeDrawdown()
        last = datas[0].tail(1)
        for index, order in self.orderBook.loc[(lambda df: (df['Stoploss'].notnull())
                                                & (df['StopExecuted'] == False)
                                                & (df['ProfitExecuted'] == False))].iterrows():
            if order.Type == 'buy' and order.Stoploss > last['low'][0]:
                self.broker.sell(
                    quantity=order.Quantity,
                    leverage=order.Leverage,
                    strategy=order.Strategy
                    )
                self.orderBook.at[index, 'StopExecuted'] = True
            elif (order.Type == 'sell' and
                  order.Stoploss < last['high'][0]):
                self.broker.buy(
                    quantity=order.Quantity,
                    leverage=order.Leverage,
                    strategy=order.Strategy
                    )
                self.orderBook.at[index, 'StopExecuted'] = True
        for index, order in self.orderBook.loc[(lambda df: (df['Takeprofit'].notnull())
                                                & (df['StopExecuted'] == False)
                                                & (df['ProfitExecuted'] == False))].iterrows():
            if order.Type == 'buy':
                if order.Takeprofit < last['high'][0]:
                    self.broker.sell(
                        quantity=order.Quantity,
                        leverage=order.Leverage,
                        strategy=order.Strategy
                    )
                    self.orderBook.at[index, 'ProfitExecuted'] = True
            if order.Type == 'sell':
                if order.Takeprofit > last['low'][0]:
                    self.broker.buy(
                        quantity=order.Quantity,
                        leverage=order.Leverage,
                        strategy=order.Strategy
                    )
                    self.orderBook.at[index, 'ProfitExecuted'] = True
        self.execute(datas)

    def computeDrawdown(self):
        balance = self.broker.balance
        if balance > self.highest:
            self.highest = balance
            self.drawdown = 0
        else:
            self.drawdown = self.highest - balance
            if balance != 0 and self.drawdown/balance > self.maxdrawdown:
                self.maxdrawdown = self.drawdown/balance

    def notifyOrder(self, order, price):
        """
        This method is called whenever the order provided by strategy is executed.
        """
        self.orderBook = self.orderBook.append({
            'Type': order.Type,
            'Limit': order.Limit,
            'Quantity': order.Quantity,
            'Leverage': order.Leverage,
            'Stoploss': order.Stoploss,
            'Takeprofit': order.Takeprofit,
            'Parent': order.Parent,
            'Strategy': order.Strategy,
            'StopExecuted': False,
            'ProfitExecuted': False,
        }, ignore_index=True)

        if order.Type == 'buy':
            self.position += order.Quantity
        elif order.Type == 'sell':
            self.position -= order.Quantity

        self.broker.notifier.notify(f"{price.index[0]}: {order.Type} "
                                    f" Executed at price {price['close'][0]:.2f}"
                                    f" Quantity: {order.Quantity}"
        )

    @abstractmethod
    def stop(self):
        """
        This method is called only once, when there is no new price bars left.
        """
        pass

    def addBroker(self, broker):
        self.broker = broker

    def addData(self, data):
        self.datas.append(data)



import sys
import asyncio


def background(f):
    def wrapped(*args, **kwargs):
        return asyncio.get_event_loop().run_in_executor(None, f,
                                                        *args, **kwargs)

    return wrapped


class Bot:

    def __init__(self):
        self.strategy = []
        self.datas = []
        self.dataIndex = 0

    def run(self):
        priceGens = []
        for data in self.datas:
            priceGens.append(data.dataGenerator())
        loop = True
        eventLoop = asyncio.get_event_loop()
        while loop:
            prices = []
            try:
                for gen in priceGens:
                    prices.append(gen.__anext__())
                prices = eventLoop.run_until_complete(asyncio.gather(*prices))
            except StopAsyncIteration:
                break
            for strategy in self.strategy:
                stratPrices = []
                for index in strategy.dataIndex:
                    stratPrices.append(prices[index])
                try:
                    strategy.broker.processData(stratPrices)
                    strategy.processData(stratPrices)
                except ValueError as VErr:
                    strategy.broker.notifier.notify(str(VErr))
                    loop = False
                    break
                except Exception as exception:
                    strategy.broker.notifier.notify(str(exception))
                    raise exception
            sys.stdout.flush()
        #eventLoop.close()
        for strategy in self.strategy:
            strategy.stop()

    def addStrategy(self, strategy):
        if strategy.broker is None:
            raise ValueError("Add Broker to Strategy first")
        self.strategy.append(strategy)
        for data in strategy.datas:
            try:
                strategy.dataIndex.append(self.datas.index(data))
            except ValueError:
                self.datas.append(data)
                strategy.dataIndex.append(self.dataIndex)
                self.dataIndex += 1



from abc import ABCMeta, abstractmethod
from .Notifier import Notifier


class Broker(metaclass=ABCMeta):

    def __init__(self, silent=False):
        self.notifier = Notifier(silent)

    @abstractmethod
    def processData(self, priceData):
        """
        This method is called every time a new price bar is recieved.
        Input priceData is list of dataframes containing price information.
        """
        pass

    @property
    @abstractmethod
    def balance(self):
        pass

    @property
    @abstractmethod
    def cashBalance(self):
        pass

    @property
    @abstractmethod
    def position(self):
        pass

    @abstractmethod
    def buy(self, quantity, leverage, stoploss=None, takeprofit=None,
            limit=None):
        pass

    @abstractmethod
    def sell(self, quantity, leverage, stoploss=None, takeprofit=None,
             limit=None):
        pass

    @abstractmethod
    def closeAllLongs(self):
        pass

    @abstractmethod
    def closeAllShorts(self):
        pass



from .Data import Data
import pandas
import datetime as dt


class CSVData(Data):

    def __init__(
            self,
            filename,
            window,
            datetime='datetime',
            volume='volume',
            open_='open',
            high='high',
            low='low',
            close='close',
            start=None,
            end=None,
            interval=60,
            freq=None,
            **params,
    ):
        fileData = pandas.read_csv(filename)
        # convert the column (it's a string) to datetime type
        datetime_series = pandas.to_datetime(fileData[datetime])

        # create datetime index passing the datetime series
        datetime_index = pandas.DatetimeIndex(datetime_series.values)
        fileData = fileData.set_index(datetime_index)
        fileData.drop(datetime, axis=1, inplace=True)
        if freq is not None:
            closeData = fileData[close].resample(freq).last()
            openData = fileData[open_].resample(freq).first()
            lowData = fileData[low].resample(freq).min()
            highData = fileData[high].resample(freq).max()
            volumeData = fileData[volume].resample(freq).sum()
            self.ohlc = pandas.concat([
                closeData,
                openData,
                lowData,
                highData,
                volumeData,
            ], axis=1).rename(columns={
                open_: 'open',
                high: 'high',
                low: 'low',
                close: 'close',
                volume: 'volume',
            })
        else:
            self.checkKey(fileData, open_)
            self.checkKey(fileData, high)
            self.checkKey(fileData, low)
            self.checkKey(fileData, close)
            self.checkKey(fileData, volume)
            self.ohlc = fileData.rename(columns={
                open_: 'open',
                high: 'high',
                low: 'low',
                close: 'close',
                volume: 'volume',
            })
        self.window = window
        self.interval = interval
        if end is None:
            self.start = fileData.index[0] + self.window
            self.end = fileData.index[-1]
        else:
            self.start = start
            assert(end is not None)
            self.end = end
        super().__init__(**params)

    async def dataGenerator(self):
        for seconds in range(0, round((self.end-self.start).total_seconds()),
                             self.interval):
            end = self.start+dt.timedelta(seconds=seconds)
            start = end-self.window
            yield self.ohlc[start:end]

    @staticmethod
    def checkKey(df, key):
        if key not in df.columns:
            raise IndexError(f"{key} column is not present in data frame")



import asyncio
import time
import pandas as pd
from .Data import Data

from bfxapi import Client


class BitfinexData(Data):

    def __init__(self, symbol='tBTCUSD', timeframe='1m', start=0, **params):
        self.bfx = Client(
            logLevel='DEBUG',
        )
        switcher = {
            '1m': 1*60,
            '5m': 5*60,
            '15m': 15*60,
            '30m': 30*60,
            '1h': 1*60*60,
            '3h': 3*60*60,
            '6h': 6*60*60,
            '12h': 12*60*60,
            '1D': 1*24*60*60,
            '7D': 7*24*60*60,
            '14D': 14*24*60*60,
            '1M': 1*30*24*60*60,
            }
        self.timeframe = timeframe
        self.interval = switcher.get(timeframe, 60)
        self.start = start
        self.symbol = symbol
        self.ohlc = pd.DataFrame(columns=[
            'open',
            'high',
            'low',
            'close',
            'volume',
        ], index=pd.to_datetime([]))
        super().__init__(**params)

    async def dataGenerator(self):
        now = int(round(time.time() * 1000))
        wait = 60
        maxRows = 2880
        blockSize = 1440
        retry = 0
        timer = None
        while retry < 10:
            if timer is not None:
                await timer

            try:
                candles = await self.bfx.rest.get_public_candles(
                    symbol=self.symbol,
                    start=self.start,
                    end=now,
                    tf=self.timeframe,
                )
            except Exception:
                await asyncio.sleep(wait)
                retry += 1
                continue

            retry = 0
            for c in candles:
                self.ohlc.loc[pd.to_datetime(c[0], unit='ms').floor('s')] = [
                    c[1],
                    c[3],
                    c[4],
                    c[2],
                    c[5],
                ]
            self.ohlc.sort_index(inplace=True)
            self.start = now
            if len(self.ohlc) >= maxRows:
                self.ohlc.drop(self.ohlc.head(blockSize).index,
                               inplace=True)
            timer = asyncio.create_task(asyncio.sleep(self.interval))
            yield self.ohlc
        raise ConnectionError("Retry Exceeded 10")

    def test(self):
        async def BitfinexDataTest(bfx, symbol, start, timeframe):
            ohlc = pd.DataFrame(columns=[
                'open',
                'high',
                'low',
                'close',
                'volume',
            ], index=pd.to_datetime([]))
            now = int(round(time.time() * 1000))
            candles = await bfx.rest.get_public_candles(
                symbol=symbol,
                start=start,
                end=now,
                tf=timeframe,
            )
            for c in candles:
                ohlc.loc[pd.to_datetime(c[0], unit='ms').floor('s')] = [
                    c[1],
                    c[3],
                    c[4],
                    c[2],
                    c[5],
                ]
            ohlc.sort_index(inplace=True)
            return ohlc

        t = asyncio.ensure_future(BitfinexDataTest(
            self.bfx,
            self.symbol,
            self.start,
            self.timeframe)
        )
        return asyncio.get_event_loop().run_until_complete(t)



import aiohttp
import asyncio
from .Data import Data


class LNMData(Data):

    def __init__(self, **params):
        self.session = aiohttp.ClientSession()
        self.wsurl = 'wss://api.lnmarkets.com/realtime'
        super().__init__(**params)

    async def dataGenerator(self):
        async with self.session.ws_connect(self.wsurl) as ws:
            async for msg in ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    if msg.data == 'close cmd':
                        await ws.close()
                        break
                    else:
                        print(msg.data)
                        await ws.send_str(msg.data + '/answer')
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    break



import pandas as pd
import random
from collections import namedtuple
from collections import deque
from .Broker import Broker

Order = namedtuple('Order', [
    'Type',
    'Limit',
    'Quantity',
    'Leverage',
    'Stoploss',
    'Takeprofit',
    'Parent',
    'Strategy',
])


class BacktestBroker(Broker):

    def __init__(self, balance, maxLeverage=1, interest=0, commission=0.001,
                 liveTrading=False, symbol='BTC'):
        self._balance = balance
        self._cashBalance = balance
        self._borrowed = 0
        self.maxLeverage = maxLeverage
        self.interest = interest
        self.commission = commission
        self.liveTrading = liveTrading
        self.symbol = symbol
        self.price = None
        self.transactions = pd.DataFrame(columns=[
            'amount',
            'price',
            'symbol',
        ], index=pd.to_datetime([]))
        self.book = pd.DataFrame(columns=[
            'balance',
            'cashBalance',
            'borrowed',
            symbol,
            'returns',
        ], index=pd.to_datetime([]))
        self.orders = deque()
        super().__init__()

    @property
    def balance(self):
        if self.price is not None:
            return (self.cashBalance
                    + self.position * self.price['close'][0]
                    - self.borrowed
                    )
        else:
            return self.cashBalance

    @property
    def cashBalance(self):
        return self._cashBalance

    @property
    def position(self):
        return self.transactions['amount'].sum()

    @property
    def borrowed(self):
        return self._borrowed

    def buy(self, strategy, quantity, leverage, stoploss=None, takeprofit=None,
            limit=None):
        self.orders.append(Order(
            Type='buy',
            Quantity=quantity,
            Leverage=leverage,
            Stoploss=stoploss,
            Takeprofit=takeprofit,
            Limit=limit,
            Parent=None,
            Strategy=strategy,
        ))

    def sell(self, strategy, quantity, leverage, stoploss=None, takeprofit=None,
             limit=None):
        self.orders.append(Order(
            Type='sell',
            Quantity=quantity,
            Leverage=leverage,
            Stoploss=stoploss,
            Takeprofit=takeprofit,
            Limit=limit,
            Parent=None,
            Strategy=strategy,            
        ))

    def closeAllLongs(self, strategy):
        if self.position > 0:
            self.orders.append(Order(
                Type='sell',
                Quantity=self.position,
                Leverage=1,
                Stoploss=None,
                Takeprofit=None,
                Limit=None,
                Parent=None,
                Strategy=strategy,
            ))

    def closeAllShorts(self, strategy):
        if self.position < 0:
            self.orders.append(Order(
                Type='buy',
                Quantity=(-self.position),
                Leverage=1,
                Stoploss=None,
                Takeprofit=None,
                Limit=None,
                Parent=None,
                Strategy=strategy,
            ))

    def calculateDebt(self, freeCapital):
        if self.borrowed >= freeCapital:
            self._borrowed -= freeCapital
        else:
            self._cashBalance = (
                self.cashBalance
                + freeCapital
                - self.borrowed
            )
            self._borrowed = 0

    def processBuy(self, order, price):
        assert(order.Leverage > 0)
        assert(order.Type == 'buy')

        cost = (1+self.commission)*order.Quantity*price/order.Leverage

        if cost > self.cashBalance:
            raise ValueError(f"Balance {self.cashBalance:.2f} not enough"
                             f" to cover cost {cost:.2f}")

        self._cashBalance = self.cashBalance - cost
        self._borrowed += cost*order.Leverage - cost

        # Add transaction
        if self.price.index[0] in self.transactions.index:
            self.transactions.loc[self.price.index[0]] = [
                (self.transactions.loc[self.price.index[0]].amount
                 + order.Quantity),
                price,
                self.symbol,
            ]
        else:
            self.transactions.loc[self.price.index[0]] = [
                order.Quantity,
                price,
                self.symbol,
            ]
        self.transactions = self.transactions.sort_index()
        order.Strategy.notifyOrder(order, self.price)

    def processSell(self, order, price):
        assert(order.Leverage > 0)
        assert(order.Type == 'sell')

        netQuantity = self.position - order.Quantity
        value = abs(netQuantity) * price
        if netQuantity < 0 and value > self.cashBalance * order.Leverage:
            raise ValueError(f"Cash Balance {self.cashBalance:.2f} not enough"
                             f" to borrow {order.Quantity} shares"
                             f" at leverage {order.Leverage:.2f}")
        freeCapital = (1-self.commission) * order.Quantity * price
        self.calculateDebt(freeCapital)

        # Add transaction
        if self.price.index[0] in self.transactions.index:
            self.transactions.loc[self.price.index[0]] = [
                (self.transactions.loc[self.price.index[0]].amount
                 - order.Quantity),
                price,
                self.symbol,
            ]
        else:
            self.transactions.loc[self.price.index[0]] = [
                - order.Quantity,
                price,
                self.symbol,
            ]
        self.transactions = self.transactions.sort_index()
        order.Strategy.notifyOrder(order, self.price)

    def processData(self, priceData):
        last = priceData[0].tail(1)
        self.price = last
        if self.balance <= 0:
            raise ValueError(
                f"{self.price.index[0]}: Account liquidated\n"
                f"Borrowed: {self.borrowed:.2f}\n"
                f"cash: {self.cashBalance:.2f}\n"
                f"Position: {self.position}, Price: {self.price['close'][0]}"
            )
        if (last.index[0].second == 0 and
            last.index[0].minute == 0 and
            last.index[0].hour == 0):
            self._cashBalance -= self.interest*self.borrowed
            try:
                self.book.loc[last.index[0]] = [
                    self.balance,
                    self.cashBalance,
                    self.borrowed,
                    self.position,
                    (self.balance - self.book.tail(1)['balance'][0])/self.balance,
                ]
            except IndexError:
                self.book.loc[last.index[0]] = [
                    self.balance,
                    self.cashBalance,
                    self.borrowed,
                    self.position,
                    0.0,
                ]

        leftOrders = deque()
        while self.orders:
            order = self.orders.popleft()
            if order.Limit is None:
                # Oder executed at random price between high and low
                weight = (random.randint(0, 100)/100)
                diff = last['high'][0]-last['low'][0]
                price = last['low'][0] + weight*diff
                if order.Type == 'buy':
                    self.processBuy(order, price)
                else:
                    self.processSell(order, price)

            else:
                if order.Type == 'buy':
                    if order.Limit > last['low'][0]:
                        self.processBuy(order, order.Limit)
                    else:
                        leftOrders.append(order)
                if order.Type == 'sell':
                    if order.Limit < last['high'][0]:
                        self.processSell(order, order.Limit)
                    else:
                        leftOrders.append(order)
        self.orders = leftOrders





import os
import asyncio
from bfxapi import Client, Order
from .Broker import Broker
from .Notifier import addMessage


class BitfinexMBroker(Broker):

    def __init__(self):
        self.bfx = Client(
            API_KEY=os.getenv("BFX_KEY"),
            API_SECRET=os.getenv("BFX_SECRET"),
            logLevel='DEBUG',
        )

    @addMessage
    def order(self, symbol, price, quantity, leverage, stoploss=None):
        async def create_order(bfx, symbol, price, quantity, leverage,
                               stoploss=None):
            if stoploss is None:
                response = await bfx.rest.submit_order(
                    symbol=symbol,
                    amount=quantity,
                    leverage=leverage,
                    price=price,
                )
            else:
                response = await bfx.rest.submit_order(
                    symbol=symbol,
                    amount=quantity,
                    leverage=leverage,
                    price=price,
                    market_type=Order.Type.STOP_LIMIT,
                    oco_stop_price=stoploss,
                    oco=True,
                )
                # response is in the form of a Notification object
            return response

        t = asyncio.ensure_future(create_order(
            bfx=self.bfx,
            symbol=symbol,
            quantity=quantity,
            leverage=leverage,
            price=price,
            stoploss=stoploss
        ))
        response = asyncio.get_event_loop().run_until_complete(t)
        for o in response.notify_info:
            self.notifier.notify('Order: '+o)



import asyncio
import time
from .Data import Data
import krakenex
from pykrakenapi import KrakenAPI


class KrakenData(Data):

    def __init__(self, interval=1, **params):
        self.KAPI = KrakenAPI(krakenex.API())
        self.ohlc, self.since = self.KAPI.get_ohlc_data(
            "BTCUSD",
            interval=interval,
            since=None,
            ascending=True,
        )
        self.interval = interval
        super().__init__(**params)

    async def dataGenerator(self):
        wait = 60
        maxRows = 2880
        blockSize = 1440
        retry = 0
        timer = None
        while retry < 10:
            if timer is not None:
                await timer
            try:
                ohlcnew, self.since = self.KAPI.get_ohlc_data(
                    "BTCUSD",
                    interval=self.interval,
                    since=self.since,
                    ascending=True,
                )
            except Exception:
                await asyncio.sleep(wait)
                retry += 1
                continue
            # Remove the last enty as it is unconfirmed
            retry = 0
            self.ohlc.drop(self.ohlc.tail(1).index, inplace=True)
            self.ohlc = self.ohlc.append(ohlcnew)
            if len(self.ohlc) >= maxRows:
                self.ohlc.drop(self.ohlc.head(blockSize).index,
                               inplace=True)
            timer = asyncio.create_task(asyncio.sleep(self.interval*wait))
            yield self.ohlc
        raise ConnectionError("Retry Exceeded 10")



import aiohttp
import asyncio
import requests
from .Data import Data


class BinanceData(Data):

    def __init__(self, window=1000, interval='1m', **params):
        self.klineurl = '/api/v3/klines'
        self.baseurl = 'https://api2.binance.com'
        self.window = window
        self.interest = interval

        params = {
            'symbol': 'BTCUSD',
            'interval': interval,
            }
        data = requests.get(self.baseurl+self.klineurl, params=params)
        self.since = data[-2][0]
        self.ohlc = self.convertData(data)
